/*
 *	DESCRIPTION:	Codeset conversion for narrow character sets.
 */


#include "cs_example.h"


/*
 * CV_convert_init
 *
 * Generic initialization function for a conversion object.
 *
 * This object converts from FROM_CS to TO_CS, using function cvt_FN and
 * the two datatables givesn.
 */
void	CV_convert_init (
CSCONVERT	csptr,
SSHORT		to_cs,
SSHORT		from_cs,
FPTR_SHORT	cvt_fn,
void		*datatable,
void		*datatable2)
{
csptr->csconvert_version = 40;
csptr->csconvert_name = (ASCII *) "DIRECT";
csptr->csconvert_from = from_cs;
csptr->csconvert_to   = to_cs;
csptr->csconvert_convert = cvt_fn;
csptr->csconvert_datatable = datatable;
csptr->csconvert_misc = datatable2;
}


/* Conversion functions
 *
 * All InterBase conversion functions follow the same API.
 *
 * The same code can be used for several character sets, as the conversion is
 * data table driven.
 *
 *
 * Note:
 *      If (dest_ptr) is NULL, then instead of performing a conversion,
 *      estimate how much space would be required for an input string of
 *      src_len bytes.
 *
 *  Input:
 *      obj         The conversion object
 *      dest_ptr    Place to put the converted string.
 *      dest_len    Length of the dest_ptr buffer, in bytes
 *      str_ptr     The input string
 *      src_len     Length of input string, in bytes.
 *
 *  Output:
 *      err_code        0, if no error occured.
                        If an error occured, a code indicating the error.
 *                      CS_CONVERT_ERROR    - input character does not exist
 *                                          in output character set.
 *                      CS_BAD_INPUT        - input charactacter string is
 *                                          illegal per rules of charset.
 *                      CS_TRUNCATION_ERROR - Not enough space in output
 *                                          buffer.
 *      err_position    Byte position in the input string where the error
 *                      occured.
 *
 *   Returns:
 *                  The length of the converted string, in bytes.
 */

 /* Several sample conversion functions are provided, each of which makes use
  * of different format data tables to drive the conversion.
  *
  * Unicode (WCHAR) to Narrow Char
  * WCHAR to WCHAR
  * Narrow Char to Unicode (WCHAR)
  * Narrow Char to Narrow Char
  */

/* CV_unicode_to_NC
 *
 * This is a generic function for converting from UNICODE in WCHAR format
 * to a target character set, which must be a NCHAR character set.
 *
 * There are two datatables, one is indexed by the high-byte of the Unicode
 * character (which defines the page) and gives the index of the conversion
 * table to use for the lowerbyte in the other table.
 */
USHORT	CV_unicode_to_nc (obj, dest_ptr, dest_len, src_ptr, src_len, err_code, err_position)
CSCONVERT	obj;
BYTE		*dest_ptr;
USHORT		dest_len;
BYTE		*src_ptr;
USHORT		src_len;
SSHORT		*err_code;
USHORT		*err_position;
{
BYTE	*start;
USHORT	src_start = src_len;
UCHAR	ch;
UNICODE	uni;

assert(	src_ptr  != NULL || dest_ptr == NULL );
assert( err_code != NULL );
assert( err_position != NULL );
assert( obj != NULL );
assert( obj->csconvert_convert == CV_unicode_to_nc );
assert( obj->csconvert_datatable != NULL );
assert( obj->csconvert_misc != NULL );

*err_code = 0;

/* See if we're only after a length estimate */
if (dest_ptr == NULL)
	return( (USHORT) ((src_len + 1) / 2) );

start = dest_ptr;
while ((src_len > 1) && dest_len)
	{
	uni = *((UNICODE *) src_ptr);
	ch = obj->csconvert_datatable[
		((USHORT *)obj->csconvert_misc)[ (USHORT) uni / 256 ]
				+ (uni % 256) ];
	if ((ch == CS_CANT_MAP) && !(uni == CS_CANT_MAP))
		{
		*err_code = CS_CONVERT_ERROR;
		break;
		}
	*dest_ptr++ = ch;
	src_ptr += 2;
	src_len -= (USHORT) 2;
	dest_len -= (USHORT) 1;
	};
if (src_len && !*err_code)
	{
	if (src_len == 1)
		*err_code = CS_BAD_INPUT;
	else
		*err_code = CS_TRUNCATION_ERROR;
	};
*err_position = (USHORT) (src_start - src_len);
return( (USHORT) (dest_ptr - start) );
}

/*
 * CV_wc_to_wc
 *
 * Generic function for conversion between two character sets in WCHAR
 * format.
 */
USHORT	CV_wc_to_wc( obj, dest_ptr, dest_len, src_ptr, src_len, err_code, err_position )
CSCONVERT	obj;
USHORT		*dest_ptr;
USHORT		dest_len;
USHORT		*src_ptr;
USHORT		src_len;
SSHORT		*err_code;
USHORT		*err_position;
{
USHORT	*start;
USHORT	ch;
UNICODE	uni;
USHORT	src_start = src_len;

assert(	src_ptr  != NULL || dest_ptr == NULL );
assert( err_code != NULL );
assert( err_position != NULL );
assert( obj != NULL );
assert( obj->csconvert_convert == CV_wc_to_wc );
assert( obj->csconvert_datatable != NULL );
assert( obj->csconvert_misc != NULL );

*err_code = 0;

/* See if we're only after a length estimate */
if (dest_ptr == NULL)
	return( src_len );

start = dest_ptr;
while ((src_len > 1) && (dest_len > 1))
	{
	uni = *((UNICODE *) src_ptr);
	ch = ((USHORT *)obj->csconvert_datatable)[
		((USHORT *)obj->csconvert_misc)[ (USHORT) uni / 256 ]
				+ (uni % 256) ];
	if ((ch == CS_CANT_MAP) && !(uni == CS_CANT_MAP))
		{
		*err_code = CS_CONVERT_ERROR;
		break;
		}
	*dest_ptr++ = ch;
	src_ptr++;
	src_len -= (USHORT) 2;
	dest_len -= (USHORT) 2;
	};
if (src_len && !*err_code)
	{
	if (src_len == 1)
		*err_code = CS_BAD_INPUT;
	else
		*err_code = CS_TRUNCATION_ERROR;
	};
*err_position = (USHORT) (src_start - src_len);
return( (USHORT) ((dest_ptr - start)*sizeof( *dest_ptr )) );
}

/*
 * CV_nc_to_unicode
 *
 * Generic function for converting a narrow character set to Unicode (WCHAR)
 */
USHORT	CV_nc_to_unicode( obj, dest_ptr, dest_len, src_ptr, src_len, err_code, err_position )
CSCONVERT	obj;
BYTE		*dest_ptr;
USHORT		dest_len;
BYTE		*src_ptr;
USHORT		src_len;
SSHORT		*err_code;
USHORT		*err_position;
{
BYTE	*start;
UNICODE	ch;
USHORT	src_start = src_len;

assert(	src_ptr  != NULL || dest_ptr == NULL );
assert( err_code != NULL );
assert( err_position != NULL );
assert( obj != NULL );
assert( obj->csconvert_convert == CV_nc_to_unicode );
assert( obj->csconvert_datatable != NULL );
assert( sizeof( UNICODE ) == 2 );

*err_code = 0;

/* See if we're only after a length estimate */
if (dest_ptr == NULL)
	return( (USHORT) (src_len * 2) );

start = dest_ptr;
while (src_len && (dest_len > 1))
	{
	ch = ((UNICODE *)(obj->csconvert_datatable))[ *src_ptr ];
	/* No need to check for CS_CONVERT_ERROR, all charsets
	 * must convert to unicode.
	 */

	*((UNICODE *)dest_ptr) = ch;
	src_ptr++;
	src_len--;
	dest_len -= (USHORT) sizeof( UNICODE );
	dest_ptr += (USHORT) sizeof( UNICODE );
	};
if (src_len && !*err_code)
	{
	*err_code = CS_TRUNCATION_ERROR;
	};
*err_position = (USHORT) (src_start - src_len);
return( (USHORT) (dest_ptr - start) );
}


/* eight_bit_convert
 *
 * Function for converting from one narrow character set to another.
 */
static USHORT	eight_bit_convert( obj, dest_ptr, dest_len, src_ptr, src_len, err_code, err_position )
CSCONVERT	obj;
BYTE		*dest_ptr;
USHORT		dest_len;
BYTE		*src_ptr;
USHORT		src_len;
SSHORT		*err_code;
USHORT		*err_position;
{
BYTE	*start;
UCHAR	ch;
USHORT	src_start = src_len;

assert(	src_ptr  != NULL || dest_ptr == NULL );
assert( err_code != NULL );
assert( err_position != NULL );
assert( obj != NULL );
assert( obj->csconvert_convert == eight_bit_convert );
assert( obj->csconvert_datatable != NULL );

*err_code = 0;

/* See if we're only after a length estimate */
if (dest_ptr == NULL)
	return( src_len );

start = dest_ptr;
while (src_len && dest_len)
	{
	ch = obj->csconvert_datatable[ *src_ptr ];
	if ((ch == CS_CANT_MAP) && (*src_ptr != CS_CANT_MAP))
		{
		*err_code = CS_CONVERT_ERROR;
		break;
		};
	*dest_ptr++ = ch;
	src_ptr++;
	src_len--;
	dest_len--;
	};
if (src_len && !*err_code)
	{
	*err_code = CS_TRUNCATION_ERROR;
	};
*err_position = (USHORT) (src_start - src_len);
return( (USHORT)(dest_ptr - start) );
}



/* Example Initiation function.
 *
 * This function initializes an object that performs conversion between
 * DOS865 and ISO Latin 1.
 */
USHORT CV_dos_865_x_latin1 (
CSCONVERT	csptr,
SSHORT		dest_cs,
SSHORT		source_cs)
{
#include "tx865_lat1.h"
if (dest_cs == CS_LATIN1)
    {
    assert (source_cs == CS_DOS_865);
	CV_convert_init( csptr, dest_cs, source_cs, eight_bit_convert,
				(void *)cvt_865_to_Latin1, NULL);
    }
else
    {
    assert (source_cs == CS_LATIN1 && dest_cs == CS_DOS_865);
	CV_convert_init( csptr, dest_cs, source_cs, eight_bit_convert,
				(void *) cvt_Latin1_to_865, NULL);
    }
return 0;
}

